import Maybe
import List		(sort,isPrefixOf,isSuffixOf,intersperse)
import Monad		(when)
import System
import Char		(isDigit,digitToInt,isUpper,toUpper,isAlphaNum)
import IO		(hFlush,stdout,hSetBuffering,BufferMode(..))
import FFIExtensions	(withCString)

import Observe		(ObserveResult(..)
			,isFound,isNotFound,isInterrupted,fromFound
			,newObservation,newObservationSrc,observableInfo)
import LowLevel		(FileNode(..),nil,openHatFile,getResult,hatVersionNumber
			,getSrcRef,srcRefFile,srcRefLine,srcRefCol
			,getDefnRef,defnFile,defnLine,defnCol)
import SExp		(SExp(..),Label,prettyEquation,prettyExpression
			,fileNode2SExp,funId,label)
import HighlightStyle	(cursorUp,cleareol,highlight,Highlight(..)
			,getTerminalSize)
import Trie		(Trie,Search(..),emptyTrie,match)
import TExp		(linearise)
import Pattern		(matchPat,parsePat,lexPat)


spawnDetectCmd = "xterm -e hat-detect "
spawnDetectEnd = "&"
spawnJavaCmd   = "hat-trail-in-java "
spawnJavaEnd   = "&"
spawnTrailCmd  = "xterm -e hat-trail "
spawnTrailEnd  = "&"

hatView :: FilePath -> Int -> Int -> String
hatView f y x  = "xterm -e hat-view "++f++" "++show y++" "++show x++" &"


shortHelpText  = ":h for help, :q to quit"

----
data InteractiveState =
    State
	{ lastObserved	:: [FileNode]
	, more		:: Bool
	, equationsPerPage :: Int
	, hatFile       :: String
	, currentPos	:: Int
	, cutoffdepth	:: Int
	, unevalMode	:: Bool
	, stringSugar	:: Bool
	, listSugar	:: Bool
	, recursiveMode	:: Bool
	, showRHS	:: Bool
	, filterMode    :: Keep
        , screenWidth	:: Int
	}
initialState :: InteractiveState
initialState = State
	{ lastObserved = []
	, more = False
	, equationsPerPage = 10
	, hatFile = ""
	, currentPos = 0
	, cutoffdepth = 10
	, unevalMode = False
	, stringSugar = True
	, listSugar = False
	, recursiveMode = True
	, showRHS = True
	, filterMode = Unique
        , screenWidth = 80
	}
----

-----------------------------------------------------------------------
-- misc functions
-----------------------------------------------------------------------

checkParameters :: String -> String -> Int
checkParameters l ('-':r) =
    checkParameters' l r
  where 
    checkParameters' _ [] = 0
    checkParameters' l (c:r) = if c `elem` l then checkParameters' l r else 2
checkParameters _ _ = 1

hasBadParameters :: String -> [String] -> Bool
hasBadParameters flags params =
    foldl (\y x-> x==2 || y) False (map (checkParameters flags) params)

goodParameters :: String -> [String] -> String
goodParameters flags l = concat (filter (\x-> checkParameters flags x ==0) l)

noParameters :: [String] -> [String]
noParameters l = filter (\x-> checkParameters [] x ==1) l

checkIdentifiers (ident1:"in":ident2:file:[]) = Just (ident1,Just ident2,file)
checkIdentifiers (ident1:file:[]) = Just (ident1,Nothing,file)
checkIdentifiers (file:[]) = Just ("",Nothing,file)
checkIdentifiers _ = Nothing

-----------------------------------------------------------------------
-- main entry point
-----------------------------------------------------------------------

main = do
    hSetBuffering stdout NoBuffering
    arguments <- getArgs
 -- putStrLn ("cmdline args: "++show arguments)
    if length arguments ==3  &&  head (tail arguments) == "-remote" then
          startObserve (rectify (head arguments)) False False False
                       (head (drop 2 arguments)) Nothing True
      else if length arguments == 5 && head (tail arguments) == "-remote" then
          startObserveSrc (rectify (head arguments)) (tail (tail arguments))
      else
          let identifiers = checkIdentifiers (noParameters arguments);
              options     = goodParameters "vxur" arguments in
          if isNothing identifiers || hasBadParameters "vxur" arguments then do
          --  putStrLn ("identifiers: "++show identifiers)
              putStrLn ("hasBadParameters: "
                        ++show (hasBadParameters "vxur" arguments))
	      putStrLn cmdlineHelp
	  else
             let verboseMode = 'v' `elem` options
                 recursiveMode = 'r' `elem` options
                 expertMode = 'x' `elem` options
                 (ident1,ident2,file) = fromJust identifiers in
	     startObserve (rectify file) verboseMode recursiveMode expertMode
                          ident1 ident2 False
  where
    rectify :: FilePath -> FilePath
    rectify f | ".hat" `isSuffixOf` f = f
              | otherwise = f ++ ".hat"


--startObserve :: String -> Bool -> Bool -> Bool -> String -> Maybe String
--                -> Bool -> IO ()
startObserve file verboseMode recursiveMode expertMode ident1 ident2 remote =
  do
    (width,_) <- getTerminalSize
    withCString file openHatFile
    when (remote ||  (ident1==""))
         (putStrLn ("\n        hat-observe "++hatVersionNumber
                    ++"    ("++shortHelpText++")\n"))
    let observed = if remote || (ident1/="") then
			  makeObserve (initialState { screenWidth=width
                                                    , hatFile=file })
                                      ident1 ident2 (SUnevaluated ())
		   else Found []
    if ident1=="" then do
	 dummy <- interactive (initialState {hatFile=file, screenWidth=width})
	 return ()
      else if remote then do
           let obs = fromFound observed
               hasmore = not (null obs)
	   when (hasmore==False)
                (putStrLn ("\nNo evaluated applications of \""++ident1
			   ++"\" found\n"))
	   dummy <- doCommand More (initialState
					{ lastObserved = obs
					, more = hasmore
					, screenWidth = width
					, hatFile = file })
           return ()
      else if isInterrupted observed then
	   putStrLn "{Interrupted}"
      else if isNotFound observed then
 	   putStrLn ("No match found.")
      else do showObservationList initialState 1 100000 (fromFound observed)
--            printCReductionList 100 (fromFound observed)
              return ()

startObserveSrc :: String -> [String] -> IO ()
startObserveSrc file args =
  do
    (width,_) <- getTerminalSize
    withCString file openHatFile
    putStrLn ("\n        hat-observe "++hatVersionNumber
              ++"    ("++shortHelpText++")\n")
    let observed = makeObserveSrc args
        obs = fromFound observed
    when (null obs)
         (putStrLn "\nNo evaluated applications found at source reference")
    dummy <- doCommand More (initialState
				{ lastObserved = fromFound observed
				, more = not (null obs)
				, screenWidth = width
				, hatFile = file })
    return ()

showObservation :: InteractiveState -> Int -> FileNode -> IO ()
showObservation state i node = do
  putStr (cursorUp ++ cleareol)	-- delay relies on line-buffered output
  putStrLn (display (screenWidth state) (cutoffdepth state)
                    (unevalMode state) (stringSugar state)
                    (listSugar state) node)
  where
  display | showRHS state = prettyEquation (show i++" ") ""
          | otherwise     = prettyExpression (show i++" ")


-- arguments to showObservationList are:
--   i     :: Int		numbered equation
--   max   :: Int		how many equations to show at once
--   exprs :: [FileNode]	roots of equations
-- result is number of equations that have been shown
showObservationList :: InteractiveState -> Int -> Int -> [FileNode] -> IO Int
showObservationList state _ _ [] = do putStr (cursorUp ++ cleareol); return 0
showObservationList state i 0 _  = do putStr (cursorUp ++ cleareol); return 0
showObservationList state i max (e:r) = do
    showObservation state i e
    putStrLn ("searching: (^C to interrupt)")
    count <- showObservationList state (i+1) (max-1) r
    return (count+1)

makeObserveSrc :: [String] -> ObserveResult
makeObserveSrc [mod,line,col] | all isDigit line && all isDigit col =
    newObservationSrc mod (fromJust (stringToInt line))
                          (fromJust (stringToInt col))
makeObserveSrc _  | otherwise = NotFound


makeObserve :: InteractiveState -> String -> Maybe String -> SExp ()
               -> ObserveResult
makeObserve state ident1 ident2 pat =
    let observed = newObservation ident1 ident2 (recursiveMode state) in
    if not (isFound observed) then observed
    else ( Found
         . map fromEqn
         . filter (\o-> matchPat o pat)
         . (case filterMode state of
              All -> id
              _   -> uniqueify)
         . map toSEqn
         . fromFound)
         observed


-- This is a unique-ifier which builds a Trie to compare expressions.
-- Each expression is represented as a token stream, obtained by first
-- converting the FileNode to an expression tree, then linearising the
-- tree (by pre-order traversal).  The token stream is matched against
-- the Trie, and if found, the corresponding FileNode is discarded.  If
-- not found, the representation is stored in the Trie, and the expression
-- returned.
uniqueify :: [SExp Label] -> [SExp Label]
uniqueify eqns = trymatch emptyTrie eqns
  where trymatch trie [] = []
        trymatch trie (eqn:eqns) =
            case match (linearise eqn) trie of
              Exists    -> trymatch trie eqns
              New trie' -> eqn: trymatch trie' eqns

-- For unique-ification *and* for pattern matching, we first need an
-- expression-tree representation of the equation at the file node.
-- This representation has a *very* deep cutoff value for correctness,
-- and no sugaring of subexprs to allow all forms of matching.
toSEqn :: FileNode -> SExp Label
toSEqn node =
  SEquation ("",nil)
            (fileNode2SExp 2000 False False True ("l",node))
            (fileNode2SExp 2000 False False False ("r",getResult node True))

-- After unique-ification and pattern matching, we want the original
-- file node back.
fromEqn :: SExp Label -> FileNode
fromEqn (SEquation _ lhs rhs) = snd (label lhs)



showSomeMore :: InteractiveState -> IO (Int,Bool)
showSomeMore state =
  let showNowList = drop (currentPos state) (lastObserved state);
      hasMore = not (null (drop (equationsPerPage state) showNowList))
  in do
    putStrLn ("searching: (^C to interrupt)")
    count <- showObservationList state (currentPos state + 1)
	                         (equationsPerPage state) showNowList
    return (count + currentPos state, hasMore)

interactive :: InteractiveState -> IO ()
interactive state
    | more state =
        do putStr "--more--> "
           hFlush stdout
           cmd <- getCommand
           case cmd of
             Observe _ -> interactive state{more=False}
             _ -> doCommand cmd state
    | otherwise =
        do putStr "hat-observe> "
           hFlush stdout
           cmd <- getCommand
           doCommand cmd state


getEquationNumber :: Int -> [FileNode] -> IO (Maybe FileNode)
getEquationNumber n lastObserved =
  let nodes = drop (n-1) lastObserved in
  if n>0 then
       if null nodes then do -- This test may take a while!
	    putStrLn "No equation with this number"
	    return Nothing
       else return (Just (head nodes))
  else return Nothing


----
setState :: Mode -> InteractiveState -> InteractiveState
setState (Uneval b)     state = state {unevalMode=b}
setState (Strings b)    state = state {stringSugar=b}
setState (Lists b)      state = state {listSugar=b}
setState (Recursive b)  state = state {recursiveMode=b}
setState (Filter k)     state = state {filterMode=k}
setState (CutOff n)     state = state {cutoffdepth=max n 1}
setState (Deeper n)     state = state {cutoffdepth=cutoffdepth state + n}
setState (Shallower n) state = state {cutoffdepth=max (cutoffdepth state - n) 1}
setState (Group n)      state = state {equationsPerPage=max n 1}

showState (Uneval _)    state = "  unevaluated exprs shown in full ("
                                ++ highlight [Underscore] "uneval"
                                ++ ") is "
                                ++ if unevalMode state then "on" else "off"
showState (Strings _)   state = "  string sugar ("
                                ++ highlight [Underscore] "strSugar"
                                ++ ") is "
                                ++ if stringSugar state then "on" else "off"
showState (Lists _)     state = "  list sugar ("
                                ++ highlight [Underscore] "listSugar"
                                ++ ") is "
                                ++ if listSugar state then "on" else "off"
showState (Recursive _) state = "  "++highlight [Underscore] "recursive"
                                ++ " calls are "
                                ++ (if recursiveMode state then "" else "NOT ")
                                ++ "shown"
showState (Filter _)    state = "  show ("
				++ concat
				     (intersperse ","
                                       (map keep [All,Unique,MostGeneralSoFar
				                 ,MostGeneral]))
                                ++ ") equations"
  where
    keep k | k==filterMode state = highlight [Underscore] (show k)
           | otherwise           = show k
showState (CutOff _) state    = "  expression "
                                ++ highlight [Underscore] "cutoff"
                                ++ " depth = "++ show (cutoffdepth state)
showState (Deeper _) state    = showState (CutOff 0) state
showState (Shallower _) state = showState (CutOff 0) state
showState (Group _)  state    = "  equations per "
                                ++ highlight [Underscore] "group"
                                ++ " = "++ show (equationsPerPage state)

----
doCommand :: Cmd -> InteractiveState -> IO()
doCommand Help state = do interactiveHelp
                          interactive state
doCommand Quit state =
    if more state then interactive state{more=False}
    else return ()
doCommand More state =
    if more state then do
        putStr (cursorUp ++ cleareol)
        (newPos,newMore) <- showSomeMore state
	interactive (state {more=newMore,currentPos=newPos})
    else do
        when (currentPos state>0) (putStrLn "No more applications observed.")
        interactive state
doCommand Info state =
     do observableInfo
	interactive state
doCommand Resize state =
     do (width,_) <- getTerminalSize
        interactive state {screenWidth=width}
doCommand Count state =
     do when (more state)
             (putStrLn "One moment, this may take a while...")
	putStrLn ("Number of (all,unique) matching applications: "
		  ++(show (length (lastObserved state))))
	interactive state

doCommand Status state =
     do mapM_ (\m-> putStrLn (showState m state))
              [ Uneval True, Strings True, Lists True, Recursive True
              , Filter All , Group 0, CutOff 0 ]
        interactive state
doCommand (Set mode) state =
     do let state' = setState mode state
        putStrLn (showState mode state')
        interactive state'

doCommand (StartTool tool n) state =
     do node <- getEquationNumber n (lastObserved state)
	when (isJust node)
             (startExternalTool tool (hatFile state) n (fromJust node))
	interactive state
doCommand (Source n) state =
     do node <- getEquationNumber n (lastObserved state)
        let sr = getSrcRef (fromJust node)
	when (isJust node && sr /= LowLevel.nil)
             (do System.system (hatView (srcRefFile sr)
                                        (srcRefLine sr)
                                        (srcRefCol sr))
                 return ())
	interactive state
doCommand (Definition n) state =
     do node <- getEquationNumber n (lastObserved state)
        let sr = getDefnRef (fromJust node)
	when (isJust node)
             (do System.system (hatView (defnFile sr)
                                        (defnLine sr)
                                        (defnCol sr))
                 return ())
	interactive state

doCommand (Observe s) state | null s = do
    putStrLn "\nObserve Wizard"
    putStrLn ""
    putStr "enter function to be observed: "
    hFlush stdout
    fun <- getLine
    putStrLn ""
    if (fun=="") then do
        putStrLn "nothing to be observed"
        interactive state         
      else do
        putStrLn "You can observe applications with specific arguments only."
        putStrLn "The '_' matches any subexpression."
        putStrLn "Enter specific arguments separated by spaces, or leave blank."
        putStr "arguments: "
        hFlush stdout
        arguments <- getLine
        putStrLn ""
	putStrLn "You can observe applications which result in a specific pattern."
        putStrLn "_|_ represents a blackholed result."
        putStr "result (or leave blank): "
        result <- getLine
        putStrLn ""
        putStrLn ("You can observe applications of \""++fun
                  ++"\" within the context of ")
        putStrLn "specific second function only."
        putStr "Enter second function (or leave blank): "
        ident2 <- getLine
        putStrLn ""
        rekMode <-
          if (ident2/="") then return ""
          else do
	    putStr ("Do you want to observe recursive calls of "++fun++"? Y/N:")
            hFlush stdout
	    rek <- getLine
            putStrLn ""
            if ((null rek)==False)&&((toUpper (head rek))=='N') then
                return "-r "
              else
	        return ""
        let query = "o "++rekMode++
		    (if null arguments then fun
                                       else ("("++fun++" "++arguments++")"))++
		    (if null result then "" else (" = "++result))++
		    (if null ident2 then "" else (" in "++ident2))
        putStrLn ("Your query is: "++query++"\n")
        doCommand (Observe query) state{showRHS=True}
        
doCommand (Observe s) state | not (null s) =
    let (pat, ctx) = parsePat (lexPat s) in
    case pat of
      Nothing -> do putStrLn ("Error in pattern")
                    interactive state
      Just p  -> do
        putStrLn ("searching: (^C to interrupt)")
        let fun = funId p
            newObserved = makeObserve state fun ctx p
        if isNotFound newObserved || null (fromFound newObserved)
          then do
            putStrLn (cursorUp ++ cleareol ++ "no match found")
            interactive state
	  else do
            putStr (cursorUp ++ cleareol)
            (newPos,newMore) <- showSomeMore
                                   (state { currentPos=0
                                          , showRHS=not (isCon fun)
	 			          , lastObserved=fromFound newObserved})
            interactive (state { lastObserved=fromFound newObserved
                               , more=newMore
                               , showRHS=not (isCon fun)
                               , currentPos=newPos})
doCommand (ObserveSrc ss) state =
    do putStrLn "searching (source reference): (^C to interrupt)"
       let newObserved = makeObserveSrc ss
       if isNotFound newObserved || null (fromFound newObserved)
         then do
           putStrLn (cursorUp ++ cleareol ++"no match found")
           interactive state
         else do
           putStr (cursorUp ++ cleareol)
           (newPos,newMore) <- showSomeMore
                                  (state { currentPos=0
                                         , showRHS=True
	 			         , lastObserved=fromFound newObserved})
           interactive (state {lastObserved=fromFound newObserved
                              ,more=newMore
                              ,showRHS=True
                              ,currentPos=newPos})
doCommand (Shell cmd) state =
     do err <- system cmd
        when (err/=ExitSuccess) (putStrLn "shell command exited abnormally")
        interactive state
doCommand Unknown state =
     do putStrLn ("Unknown command.  "++shortHelpText)
        interactive state


isCon :: String -> Bool
isCon (x:xs) = isUpper x || x `elem` ":[,"

startExternalTool tool file number node =
    let id = show (int node)
        rhsID = show (int (getResult node True))
    in case tool of
      Choose -> do
          putStr "start hat-detect or hat-trail? (D/T): "
          s <- getLine
          let choice = case map toUpper s of "D"->Detect; "T"->Trail;
                                             "J"->Java;   _ ->Choose;
          startExternalTool choice file number node
      Java -> do
          putStr ("Equation "++(show number)
                  ++": Trace left-hand-side (lhs) or rhs of equation? (L/R): ")
	  lhs <- getLine
          errcode <- system (spawnJavaCmd++file++" -remote "
			     ++(if rhsID=="" || length lhs==0 ||
				   toUpper (head lhs) /='R'
                                then id else rhsID)
			     ++spawnJavaEnd)
          when (errcode/=ExitSuccess)
	       (putStrLn ("ERROR: Unable to start hat-trail-in-java.\n"
		          ++"Check settings and availability."))
      Trail -> do
          errcode <- system (spawnTrailCmd++file++" -remote "++id
                             ++spawnTrailEnd)
          when (errcode/=ExitSuccess)
	       (putStrLn ("ERROR: Unable to start hat-trail.\n"
		          ++"Check settings and availability of hat-trail."))
      Detect -> do
         errcode <- system (spawnDetectCmd++file++" -remote "
                            ++id++spawnDetectEnd)
	 when (errcode/=ExitSuccess)
	      (putStrLn ("ERROR: Unable to start hat-detect.\n"
		         ++"Check settings and availability of hat-detect."))

interactiveHelp = do
    putStr helptext
    s <- getLine
    when (null s) (putStrLn queryHelp)
  where
    helptext = "\
\---------------------------------------------------------------------------\n\
\ <query>            observe the named function/pattern (syntax below)\n\
\ <RETURN>           show more observations (if available)\n\
\ :observe <query>   observe the named function/pattern (syntax below)\n\
\ :info              see a list of all observable functions\n\
\ :detect <n>        start hat-detect on equation <n>\n\
\ :trail <n>         start hat-trail browser on equation <n>\n\
\ :source <n>        show the source application for equation <n>\n\
\ :Source <n>        show the source definition for identifier in eqn <n>\n\
\ :set               show all current mode settings\n\
\ :set <flag>        change one mode setting\n\
\   <flag> can be: uneval [on|off]      show unevaluated expressions in full\n\
\                  strSugar [on|off]    sugar character strings\n\
\                  listSugar [on|off]   sugar lists\n\
\                  recursive [on|off]   show recursive calls\n\
\                  [all|unique]         show all equations or only unique\n\
\                  group <n>            number of equations listed per page\n\
\                  cutoff <n>           cut-off depth for deeply nested exprs\n\
\ :+[n]              short-cut to increase cutoff depth by <n> (default 1)\n\
\ :-[n]              short-cut to decrease cutoff depth by <n> (default 1)\n\
\ :resize            detect new window size for pretty-printing\n\
\ :help  :?          show this help text\n\
\ :quit              quit\n\
\---------------------------------------------------------------------------"

    queryHelp = "\
\                 The syntax of an :observe <query>.\n\
\ A simple function identifier finds all applications of that function.\n\
\ To restrict the number of equations, follow the function name with\n\
\ argument or result patterns.  Another way of refining the search is\n\
\ to ask for calls only from a specific enclosing function.\n\
\\n\
\ The full query syntax is:\n\
\    identifier [pattern]*  ['=' pattern]?  ['in' identifier]?\n\
\ where\n\
\    pattern = _                                wildcard\n\
\            | num                              number\n\
\            | ''' char '''                     character\n\
\            | '\"' string '\"'                   string\n\
\            | '[' pattern [',' pattern]* ']'   literal list\n\
\            | Con                              nullary constructor\n\
\            | '(' Con [pattern]* ')'           constructor application\n\
\ Infix constructors take the normal syntax, apart from associativity.\n\
\ Examples:\n\
\     myfn\n\
\     myfn _ (Con 2 _) (1 `Con` 3) in myOtherFn\n\
\     myfn \"Hello World!\" (1:(2:_)) = [1,_]\n\
\     myfn ((:) 1 ((:) 2 _)) \n\
\---------------------------------------------------------------------------"

cmdlineHelp = "\
\Usage: hat-observe prog[.hat]\n\
\         An interactive tool to show actual function applications within\n\
\         a traced run of a Haskell program."

{-
cmdlineHelp = "\
\Usage:   hat-observe [-v] [-r] [-xu] identifier [in topidentifier] filename\n\
\Description:\n\
\       prints a table of all applications and results of the given\n\
\       top-level identifier [within the application of topidentifier].\n\
\Options:\n\
\       v: verbose mode. Unevaluated expressions are shown in full.\n\
\       r: recursive mode.  Do not omit recursive function applications.\n\
\       xu: expert's mode for a very fast response. All applications\n\
\           of the identifier are shown, rather than only the most\n\
\           general ones."
-}


data Cmd  = Observe String | ObserveSrc [String]
          | More | Info | Count
          | Help | Quit | Unknown
          | StartTool Tool Int | Source Int | Definition Int
          | Status | Set Mode | Shell String | Resize
data Tool = Trail | Java | Detect  | Choose
data Mode = Uneval Bool | Strings Bool | Lists Bool | Recursive Bool
          | CutOff Int  | Deeper Int | Shallower Int
          | Group Int   | Filter Keep
data Keep = All | Unique | MostGeneralSoFar | MostGeneral deriving Eq

getCommand :: IO Cmd
getCommand = do
    s <- getLine
    if null s then return More
   -- else if all isDigit s then return (number (StartTool Choose) [s] 0)
      else if head s /= ':' then return (Observe s)
      else case words (tail s) of
          [] -> return Unknown
          (cmd:ss)
              | cmd `isPrefixOf` "quit" -> return Quit
              | cmd `isPrefixOf` "help" -> return Help
              | cmd `isPrefixOf` "observe"  ->
					return (Observe (unwords ss))
              | cmd `isPrefixOf` "location" ->
					return (ObserveSrc ss)
              | cmd `isPrefixOf` "detect"   ->
					return (number (StartTool Detect) ss 0)
              | cmd `isPrefixOf` "trail"    ->
					return (number (StartTool Trail) ss 0)
              | cmd `isPrefixOf` "java"     ->
					return (number (StartTool Java) ss 0)
              | cmd `isPrefixOf` "source"   -> return (number Source ss 0)
              | cmd `isPrefixOf` "Source"   -> return (number Definition ss 0)
              | cmd `isPrefixOf` "info"     -> return Info
              | cmd `isPrefixOf` "resize"   -> return Resize
              | cmd `isPrefixOf` "set"      ->
                  case ss of
                      [] -> return Status
                      (m:sss)
                          | m `isPrefixOf` "uneval" ->
					return (onOff Uneval sss)
                          | m `isPrefixOf` "strSugar" ->
					return (onOff Strings sss)
                          | m `isPrefixOf` "listSugar" ->
					return (onOff Lists sss)
                          | m `isPrefixOf` "recursive" ->
					return (onOff Recursive sss)
                          | m `isPrefixOf` "all" ->
					return (Set (Filter All))
                          | m `isPrefixOf` "unique" ->
					return (Set (Filter Unique))
                          | m `isPrefixOf` "generalise" ->
					return (Set (Filter MostGeneralSoFar))
                          | m `isPrefixOf` "mostgeneral" ->
					return (Set (Filter MostGeneral))
                          | m `isPrefixOf` "cutoff" ->
					return (Set (number CutOff sss 10))
                          | m `isPrefixOf` "group" ->
					return (Set (number Group sss 10))
                          | otherwise -> return Unknown
              | head cmd == '?' -> return Help
              | head cmd == '+' ->
                          return (Set (number Deeper (tail cmd:ss) 1))
              | head cmd == '-' ->
                          return (Set (number Shallower (tail cmd:ss) 1))
              | head cmd == '!' -> return (Shell (unwords (tail cmd:ss)))
              | otherwise  ->  return Unknown

number :: (Int->a) -> [String] -> Int -> a
number cons s def = (maybe (cons def) cons . stringToInt . unwords) s

onOff :: (Bool->Mode) -> [String] -> Cmd
onOff mode s | null s = Set (mode True)
             | otherwise = case head s of "on" -> Set (mode True)
                                          "active" -> Set (mode True)
                                          "off" -> Set (mode False)
                                          "no" -> Set (mode False)
                                          _ -> Unknown

stringToInt :: String -> Maybe Int
stringToInt s = stringToInt' True 0 s
 where
  stringToInt' True _ ('#':r) = stringToInt' True 0 r -- skip "#" at beginning
  stringToInt' True _ (' ':r) = stringToInt' True 0 r -- skip " " at beginning
--stringToInt' False i (' ':r) = Just i
  stringToInt' first i [] = if first then Nothing else  Just i
  stringToInt' _ i (c:r)
      | isDigit c = stringToInt' False (i*10+(digitToInt c)) r
      | otherwise = Nothing

instance Show Keep where
  showsPrec p All		= showString "all"
  showsPrec p Unique 		= showString "unique"
  showsPrec p MostGeneralSoFar	= showString "generalise"
  showsPrec p MostGeneral	= showString "mostgeneral"
