-- hat-observe permits matching of candidate equations against a pattern.
-- This requires a parser to convert a pattern from a string into an SExp,
-- and a comparison function on SExps.
module Pattern
  ( matchPat
  , lexPat
  , parsePat
  ) where

import Prelude hiding (lex)
import List	(isPrefixOf)
import Char	(isDigit,isAlphaNum,isSpace,isUpper,isLower)
import SExp	(SExp(..),SFixity(..))
import ParseLib



-- Comparison of an expression against a pattern.
-- All SExp labels are ignored.
-- First arg is candidate, second is pattern.
matchPat :: SExp a -> SExp b -> Bool
matchPat _                 (SUnevaluated _)  = True
matchPat (SApp _ as)       (SApp _ bs)       = and (zipWith matchPat as bs)
matchPat (SId _ a _)       (SId _ b _)       = a==b
matchPat (SApp _ (a:_))  b@(SId _ _ _)       = matchPat a b
matchPat (SLiteral _ a)    (SLiteral _ b)    = a==b
matchPat (SString _ a)      b                = matchString a b
matchPat (SLambda _)       (SLambda _)       = True
matchPat (SBottom _)       (SBottom _)       = True
matchPat (SEquation _ a b) (SEquation _ c d) = matchPat a c && matchPat b d
matchPat (SEquation _ a b) p                 = matchPat a p
matchPat _                 _                 = False

-- No matches for SWithin, SIf, SCase, SGuard, SCut, because
-- you can't specify such a pattern.


-- Simple FSM lexer for patterns.
data Token
  = Variable String
  | Constructor String
  | InfixVariable String
  | InfixConstructor String
  | Numeric String
  | String String
  | Char String
  | OpenParen
  | CloseParen
  | OpenBracket
  | CloseBracket
  | Comma
  | Underscore
  | Bottom
  | Equal
  | In
  | End
  deriving (Eq,Show)

accumulateUntil :: (Char->Bool) -> (String->Token) -> (String->[Token])
                   -> String -> (String->[Token])
accumulateUntil pred tok k acc [] = tok (reverse acc): k []
accumulateUntil pred tok k acc (c:cs)
    | pred c    = tok (reverse acc): k (c:cs)
    | otherwise = accumulateUntil pred tok k (c:acc) cs

lexPat :: String -> [Token]
lexPat [] = [End]
lexPat (x:xs)
  | x=='('     = OpenParen:    lexPat xs
  | x==')'     = CloseParen:   lexPat xs
  | x=='['     = OpenBracket:  lexPat xs
  | x==']'     = CloseBracket: lexPat xs
  | x==','     = Comma:        lexPat xs
  | x=='_' && "|_" `isPrefixOf` xs
               = Bottom:       lexPat (drop 2 xs)
  | x=='_'     = Underscore:   lexPat xs
  | x=='=' && (null xs || not (isSymbol (head xs)))
               = Equal:        lexPat xs
  | x=='i' && "n " `isPrefixOf` xs
               = In:           lexPat (drop 2 xs)
  | isSpace x  = lexPat xs
  | isDigit x  = accumulateUntil (not.isNumeric) Numeric lexPat [x] xs
  | isUpper x  = accumulateUntil (not.isVariable) Constructor lexPat [x] xs
  | isLower x  = accumulateUntil (not.isVariable) Variable lexPat [x] xs
  | isSymbol x = accumulateUntil (not.isSymbol) InfixVariable lexPat [x] xs
  | x==':'     = accumulateUntil (not.isSymbol) InfixConstructor lexPat [x] xs
  | x=='`'     = accumulateUntil (=='`') InfixVariable (lexPat.tail) [] xs
  | x=='\''    = accumulateUntil (=='\'') Char (lexPat.tail) [] xs
  | x=='"'     = accumulateUntil (=='"') String (lexPat.tail) [] xs
  | otherwise  = [End]

isSymbol, isNumeric, isVariable :: Char -> Bool
isSymbol x   = x `elem` "!#$%&*+./<=>?@\\^|-~"
isNumeric x  = isDigit x    || x `elem` ".eE"
isVariable x = isAlphaNum x || x `elem` "_'"


-- Parser from [Token] to SExp.
parsePat :: [Token] -> (Maybe (SExp ()), Maybe String)
parsePat tokens = case papply context tokens of
                    [((exp,ctx),[End])] -> (Just exp, ctx)
                    _                   -> (Nothing, Nothing)

context  :: Parser Token (SExp (), Maybe String)
context =
    do eqn <- equation
       ctx <- ( do tok In
                   Variable ctx  <- item
                   return (Just ctx) ) +++
              ( do return Nothing )
       return (eqn,ctx)
            
equation :: Parser Token (SExp ())
equation =
    do lhs <- topPattern
       ( ( do tok Equal
              rhs <- topPattern
              return (SEquation () lhs rhs) ) +++
         ( do return lhs ) )

topPattern :: Parser Token (SExp ())
topPattern =
  ( do InfixConstructor c <- item
       return (SId () c SInfixDefault) ) +++
  ( do InfixVariable v <- item
       return (SId () v SInfixDefault) ) +++
  ( do ps <- many1 pattern
       ( case ps of
           [p] -> return p
           _   -> return (SApp () ps)) )

pattern :: Parser Token (SExp ())
pattern =
  ( do String s <- item
       return (mkList (SLiteral () . show) s) ) +++
  ( do tok Underscore
       return (SUnevaluated ()) ) +++
  ( do Char c <- item
       return (SLiteral () ('\'':c++"'")) ) +++
  ( do Numeric n <- item
       return (SLiteral () n) ) +++
  ( do Constructor c <- item
       return (SId () c SInfixDefault) ) +++
  ( do Variable v <- item
       return (SId () v SInfixDefault) ) +++
  ( do tok OpenParen
       tok CloseParen
       return (SId () "()" SInfixDefault) ) +++
  ( do tok OpenParen
       InfixVariable v <- item
       tok CloseParen
       return (SId () v SInfixDefault) ) +++
  ( do tok OpenParen
       InfixConstructor c <- item
       tok CloseParen
       return (SId () c SInfixDefault) ) +++
  ( do tok OpenParen
       Constructor c <- item
       ps <- many pattern
       tok CloseParen
       return (SApp () (SId () c SInfixDefault: ps)) ) +++
  ( do tok OpenParen
       p <- pattern
       InfixConstructor c <- item
       ps <- many pattern
       tok CloseParen
       return (SApp () (SId () c SInfixDefault: p: ps)) ) +++
  ( do tok OpenParen
       p <- pattern
       InfixVariable c <- item
       ps <- many pattern
       tok CloseParen
       return (SApp () (SId () c SInfixDefault: p: ps)) ) +++
  ( do tok OpenParen
       ps <- pattern `sepby1` (tok Comma)
       tok CloseParen
       let tuple = replicate (length ps - 1) ','
       return (SApp () (SId () tuple SInfixDefault: ps)) ) +++
  ( do tok OpenBracket
       elems <- pattern `sepby` (tok Comma)
       tok CloseBracket
       return (mkList id elems) )

mkList :: (a->SExp ()) -> [a] -> SExp ()
mkList one []     = SId () "[]" SInfixDefault
mkList one (x:xs) = SApp () [ SId () ":" SInfixDefault
                            , one x
                            , mkList one xs ]

matchString :: String -> SExp a -> Bool
matchString []     (SId _ "[]" _) = True
matchString (c:cs) (SApp _ [SId _ ":" _ , SLiteral _ c', tl]) =
        c' == show c  &&  matchString cs tl
matchString _      _ = False
