FiniteMap.hs
-- of a domain element and an associated range element
-- held in binary search tree ordered by the domain elements.

module FiniteMap (FiniteMap, emptyFM, applyFM, overFM) where

data Ord a => FiniteMap a b =
    NullFM
  | NodeFM (FiniteMap a b) a b (FiniteMap a b)

-- The empty mapping.
emptyFM :: Ord a => FiniteMap a b
emptyFM = NullFM

-- Apply a mapping as a partial function.
applyFM :: Ord a => FiniteMap a b -> a -> Maybe b
applyFM NullFM               _ = Nothing
applyFM (NodeFM fm0 p q fm1) x = case compare x p of
				 LT -> applyFM fm0 x
				 EQ -> Just q
				 GT -> applyFM fm1 x

-- Extend or revise a mapping according to a fresh pair
-- of domain and range elements.
overFM :: Ord a => FiniteMap a b -> a -> b -> FiniteMap a b
overFM NullFM               x y = NodeFM NullFM x y NullFM
overFM (NodeFM fm0 p q fm1) x y = case compare x p of
				  LT -> NodeFM (overFM fm0 x y) p q fm1
				  EQ -> NodeFM fm0 x y fm1
				  GT -> NodeFM fm0 p q (overFM fm1 x y)

-- Colin Runciman, March 2002
module Main where
import FiniteMap
import Sequence

data SLMD = Same | Less | More | Diff deriving Show

lub Diff _    = Diff
lub More Diff = Diff
lub More Less = Diff
lub More More = More
lub More Same = More
lub Less Diff = Diff
lub Less More = Diff
lub Less Less = Less
lub Less Same = Less
lub Same c    = c

data Node a = Uneval | Zero | Two a a

data Graph = Graph (Sequence (Node Int)) Int

cmpGraph :: Graph -> Graph -> SLMD
cmpGraph g1 g2 = fst (cmpGraphInCxt (Same,emptyFM) g1 g2)

type Context = (SLMD, FiniteMap (Int,Int) SLMD)

cmpGraphInCxt :: Context -> Graph -> Graph -> Context
cmpGraphInCxt (Diff,m) _            _             = (Diff,m)
cmpGraphInCxt (c,m)   (Graph m0 r0) (Graph m1 r1) =
  case applyFM m (r0,r1) of
  Nothing ->
    let (c',m) = cmpNode (indexS m0 r0) (indexS m1 r1) in
    (c', overFM m (r0,r1) c')
  Just c' ->
    (c `lub` c',m)
  where
  cmpNode Uneval     Uneval   = (c,m)
  cmpNode Uneval     _        = (c `lub` Less,m)
  cmpNode _          Uneval   = (c `lub` More,m)
  cmpNode Zero       Zero     = (c,m)
  cmpNode (Two v w) (Two x y) = cmpGraphInCxt
                                  (cmpGraphInCxt
				     (c,overFM m (r0,r1) Same)
                		     (Graph m0 v) (Graph m1 x))
        			  (Graph m0 w) (Graph m1 y)
  cmpNode _          _        = (Diff,m)

---------------8<-------Just for testing----8<---------------
  
graphA = Graph (emptyS
         `snocS` Two 1 2   -- node 0
	 `snocS` Two 0 3   -- node 1
	 `snocS` Zero      -- node 2
	 `snocS` Uneval
         ) 0

graphB = Graph (emptyS
         `snocS` Two 1 2   -- node 0
	 `snocS` Two 0 3   -- node 1
	 `snocS` Uneval    -- node 2
	 `snocS` Zero
	 ) 0

main = print (cmpGraph graphA graphB)


type Sequence a = [a]

emptyS = []

snocS xs x = xs ++ [x]

indexS xs i = xs !! i

