/* This is an implementation of a linear search through the .hat file,
 * building a list of all identifiers found, with application counts.
 */
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <glib.h>
#include "pathutils.h"
#include "art.h"
#include "artutils.h"

#define DEBUG 0
#if DEBUG
#define HIDE(x) x
#else
#define HIDE(x)
#endif

#define MAX_FILENAME 1024

/* We are interested in distinguishing three different kinds of identifier:
 *    variables defined at the top level
 *    variables defined in a local scope
 *    data constructors
 */
typedef enum {
    TopId, LocalId, Construct
} idkind;

/* There are five things we are interested in for each identifier:
 *   - its name
 *   - its kind (local, global, constructor)
 *   - its arity
 *   - how many times it was applied at its full arity and gave a result
 *   - how many times it was applied at its full arity but did not give a result
 */
typedef struct {
    char* name;
    idkind kind;
    unsigned char arity;
    int uses;		/* number of evaluated applications */
    int pending;	/* number of unevaluated applications */
} item;

typedef struct {
    FileOffset atom;
    unsigned char arity;
} defn;

/* There are two finitemap structures that associate file pointers
 * with items.
 *
 * The first maps an AtomVariable or AtomConstructor pointer
 * to the information (struct item) we are interested in.
 *
 * The second maps an ExpValueUse, ExpConstUse, or ExpConstDef pointer
 * to the Atom pointer it contains (struct defn).  It also maps any
 * undersaturated application (ExpApp) to the Atom pointer.  The (struct
 * defn) contains a note of the remaining arity so we can tell when full
 * saturation is reached.
 *
 * In the second map, each atom might be the target of several
 * trace pointers, e.g. different usage sites in the source lead to
 * different ExpValueUse nodes, each pointing to the same Atom;
 * e.g.2. an undersaturated ExpApp node (partial application) can
 * also map to an Atom.
 *
 * When searching, when we find:
 *      an Atom -> add it to the first map
 *      an ExpValueUse -> add it to the second map (no increment)
 *      an ExpConstDef -> add it to the second map (no increment)
 *      an ExpConstUse -> add it to the second map (increment item)
 *      an ExpApp -> check arity, if saturated, just increment item
 *                                otherwise add to 2nd map (no increment)
 */
GTree *map1, *map2, *globals=0, *locals=0, *constrs=0;
void
map1_insert (FileOffset node, char* id, idkind k, unsigned char arity)
{
  item *it;
  it = (item*)malloc(sizeof(item));
  it->name  = id;
  it->kind  = k;
  it->arity = arity;
  it->uses  = 0;
  it->pending = 0;
  g_tree_insert(map1,(gpointer)node,(gpointer)it);
}
item *
map2_insert (FileOffset usage, FileOffset def, unsigned char ap)
{
  item *it=(item*)0;
  it = g_tree_lookup(map1,(gpointer)def);
  if (it) {
    defn *fn;
    fn = (defn*)malloc(sizeof(defn));
    fn->atom  = def;
    fn->arity = it->arity - ap;
  //if (strcmp(it->name,">=")==0)
  //  fprintf(stderr,"map2: %s at 0x%x (%d)\n",it->name,usage,it->uses);
    g_tree_insert(map2,(gpointer)usage,(gpointer)fn);
  }
  return it;
}
int
item_sort (FileOffset node, item *it, void* dummy)
{
  item *already;
  switch (it->kind) {
    case TopId:
      already = g_tree_lookup(globals,(gpointer)it->name);
      if (already) {
        it->uses += already->uses;
        it->pending += already->pending;
      }
      g_tree_insert(globals,(gpointer)it->name,(gpointer)it);
      break;
    case LocalId:
      already = g_tree_lookup(locals,(gpointer)it->name);
      if (already) {
        it->uses += already->uses;
        it->pending += already->pending;
      }
      g_tree_insert(locals, (gpointer)it->name,(gpointer)it);
      break;
    case Construct:
      already = g_tree_lookup(constrs,(gpointer)it->name);
      if (already) {
        it->uses += already->uses;
        it->pending += already->pending;
      }
      g_tree_insert(constrs,(gpointer)it->name,(gpointer)it);
      break;
    default: break;
  }
  return FALSE;
}


/* For printing the variable list with application counts, we need to
 * establish the longest name and the largest count, then columnate
 * the whole list.
 */
#define MAX_IDENT 1024
char	*idents[MAX_IDENT],    *counts[MAX_IDENT];
int	ident_size[MAX_IDENT], count_size[MAX_IDENT];
int	ident_index=0;

int
item_print (char *name, item *it, void* dummy)
{
  if (it->uses || it->pending) {
    if (ident_index >= MAX_IDENT) {
      fprintf(stderr,"out of space for identifiers\n");
      return TRUE;
    }
    idents[ident_index] = name;
    ident_size[ident_index] = strlen(name);
    counts[ident_index] = (char*)malloc(35);
    if (it->pending) {
      snprintf(counts[ident_index],34,"[1;31m%d+[0m[34m%d"
              ,it->pending,it->uses);
      count_size[ident_index] = strlen(counts[ident_index]) - 16;
    } else {
      snprintf(counts[ident_index],34,"%d",it->uses);
      count_size[ident_index] = strlen(counts[ident_index]);
    }
    ident_index++;
  }
  return FALSE;
}
void
columnate (int width)
{
  int i, j, max_count=0, max_ident=0, column_width, columns, lines, me;
  for (i=0; i<ident_index; i++) {
    if (count_size[i] > max_count) max_count=count_size[i];
    if (ident_size[i] > max_ident) max_ident=ident_size[i];
  }
  column_width = max_count + max_ident + 4;
  columns = width / column_width;
  if (ident_index % columns)
       lines = ident_index / columns + 1;
  else lines = ident_index / columns;
  for (i=0; i<lines; i++) {
    for (j=0; j<columns; j++) {
      me = j*lines + i;
      if (me < ident_index) {
        fprintf(stdout,"[34m%*s[0m %-*s"
                      ,max_count,counts[me],max_ident+3,idents[me]);
      }
    }
    fprintf(stdout,"\n");
  }
}



/* q_oneNode() moves the file pointer past a single node in the file.
 * As a side-effect, if it finds an AtomVariable or AtomConstructor,
 * it adds it to the global structure 'map1'.  If it finds an ExpValueUse
 * or ExpConstDef, it adds an entry in map2 from that usage to the relevant
 * Atom in map1.  If it finds an ExpApp or ExpConstUse, it instead looks
 * up the Atom ptr in map2, then looks up that Atom in map1, and finally
 * increments the usage counter.  However, in the case where an ExpApp
 * is undersaturated (discovered by comparing its arity with the arity
 * stored in map2), rather than incrementing the usage counter, we
 * instead need to add the address of the ExpApp to map2.
 */
void
q_oneNode (void)
{
  char c; int err;
  FileOffset node = q_position;
/*fprintf(stdout,"\n0x%x: ",position); fflush(stdout);*/
  err = q_fread(&c,sizeof(char),1,HatFileSeq);
  if (err!=1) return;
  switch (lower5(c)) {	/* lower 5 bits identify the TraceType */
    case ExpApp:
        if (hasSrcPos(c)) { q_readFO(); }
        q_readFO();		/* skip parent */
        { unsigned char size, next, i;
          FileOffset fun, result; defn *def; item *it;
          result = q_readFO();	/* get result */
          fun = q_readFO();	/* keep fun ptr */
          q_fread(&size,sizeof(unsigned char),1,HatFileSeq);	/* get arity */
          for (i=0; i<size; i++) q_readFO();	/* skip args */
          def = (defn*)g_tree_lookup(map2,(gpointer)fun);
          if (def) {
            it = g_tree_lookup(map1,(gpointer)def->atom);
            if (it) {
              if (size >= def->arity) {
                if (result==Entered) it->pending += 1;
                else if (result!=Unevaluated) it->uses += 1;
              } else {
                map2_insert(node,def->atom,size);
              }
            } else {
              fprintf(stderr,"unknown atom in fun at (ExpApp 0x%x)\n",node);
            }
          } else {
       //   fprintf(stderr,"unknown fun at (ExpApp 0x%x)\n",node); 
          }
        } break;
    case ExpValueApp:
        if (hasSrcPos(c)) { q_readFO(); }
        q_readFO();		/* skip parent */
        { unsigned char size, next, i;
          FileOffset fun; defn *def; item *it;
          fun = q_readFO();	/* fun ptr is an Atom ref */
          q_fread(&size,sizeof(unsigned char),1,HatFileSeq);	/* get arity */
          for (i=0; i<size; i++) q_readFO();	/* skip args */
          it = g_tree_lookup(map1,(gpointer)fun);
          if (it) {
            if (size >= it->arity) {
              it->uses += 1;
            } else {
              map2_insert(node,fun,size);
            }
          } else {
            fprintf(stderr,"unknown atom in fun at (ExpValueApp 0x%x)\n",node);
          }
        } break;
    case ExpValueUse:
        if (hasSrcPos(c)) { q_readFO(); }
        q_readFO();		/* skip parent */
        { FileOffset atom; item *it;
          atom = q_readFO();	/* get atom */
          if (atom!=Lambda) {
            it = g_tree_lookup(map1,(gpointer)atom);
            if (it) {
              map2_insert(node,atom,0);
              if ((it->kind==Construct) && (it->arity==0)) it->uses+=1;
            } else fprintf(stderr,"unknown atom in (ExpValueUse 0x%x)\n",node);
          }
        } break;
    case ExpConstUse:
        if (hasSrcPos(c)) { q_readFO(); }
        q_readFO();		/* skip parent */
        { FileOffset exp; defn *def; item *it;
          exp = q_readFO();	/* get ExpConstDef location */
          def = g_tree_lookup(map2,(gpointer)exp);
          if (def) {
            map2_insert(node,def->atom,0);
            it = g_tree_lookup(map1,(gpointer)def->atom);
            if (it) it->uses+=1;
            else fprintf(stderr,"unknown atom in defn in (ExpConstUse 0x%x)\n",node);
          } // else fprintf(stderr,"unknown defn in (ExpConstUse 0x%x)\n",node);
        } break;
    case ExpConstDef:
        q_readFO();		/* skip parent */
        q_readFO();		/* skip result */
        { FileOffset atom; item *it;
          atom = q_readFO();	/* get atom */
          it = g_tree_lookup(map1,(gpointer)atom);
          if (it) map2_insert(node,atom,0);
          else fprintf(stderr,"unknown atom in (ExpConstDef 0x%x)\n",node);
        } break;
    case AtomVariable:
        q_readFO();		/* skip module pointer */
        { int x; q_fread(&x,sizeof(int),1,HatFileSeq); }   /* skip line/col */
        { char x; q_fread(&x,sizeof(char),1,HatFileSeq); } /* skip fixity */
        { unsigned char arity;
          char *id;
          idkind k;
          q_fread(&arity,sizeof(unsigned char),1,HatFileSeq);
          id = q_readString();
          k = (localDef(c) ? LocalId : TopId);
          map1_insert(node,id,k,arity);
        }
        break;
    case AtomConstructor:
        q_readFO();		/* skip module pointer */
        { int x; q_fread(&x,sizeof(int),1,HatFileSeq); }   /* skip line/col */
        { char x; q_fread(&x,sizeof(char),1,HatFileSeq); } /* skip fixity */
        { unsigned char arity;
          char *id;
          q_fread(&arity,sizeof(unsigned char),1,HatFileSeq);
          id = q_readString();
          if (hasFields(c)) for (;arity-->0;) q_readFO();
          map1_insert(node,id,Construct,arity);
        }
        break;
    case AtomAbstract:
        { char* id;
          id = q_readString();
          map1_insert(node,id,Construct,0);
        }
        break;
    default:
        q_skipNode(c);
        break;
  }
}

void
observableInfo (void)
{
  int err;

  if (!globals) {
    fprintf(stdout,"Searching...\n");

    q_position = 0x10;
    fseek(HatFileSeq,q_position,SEEK_SET);

    map1 = g_tree_new((GCompareFunc)fileoffset_compare);
    map2 = g_tree_new((GCompareFunc)fileoffset_compare);
    do {
      q_oneNode();
    } while (!feof(HatFileSeq));

    globals = g_tree_new((GCompareFunc)strcmp);
    locals  = g_tree_new((GCompareFunc)strcmp);
    constrs = g_tree_new((GCompareFunc)strcmp);
    g_tree_traverse(map1,(GTraverseFunc)item_sort,G_IN_ORDER,(void*)0);

    g_tree_traverse(globals,(GTraverseFunc)item_print,G_IN_ORDER,(void*)0);
    fprintf(stdout,"[A[K");	/* cursor up and clear-to-eol */
  }

//fprintf(stdout,"[34mVariables:[0m\n");
  columnate(80);
#if 0
  fprintf(stdout,"[34mLocal Variables:[0m\n");
  g_tree_traverse(locals, (GTraverseFunc)item_print,G_IN_ORDER,(void*)0);
  fprintf(stdout,"[34mConstructors:[0m\n");
  g_tree_traverse(constrs,(GTraverseFunc)item_print,G_IN_ORDER,(void*)0);
#endif

  /* note: some data-structure clean-up needed? */
}

#if 0
int
main (int argc, char **argv)
{
  int err;
  char header[8];
  char *arg, *dir;

  progname = basename(argv[0],0);
                        /* for error messages - /not/ the prog being debugged */
  if (argc!=2) {
    fprintf(stderr,"  [%s]\n",argv[0]);
    fprintf(stderr,"\tUsage: %s program[.hat]\n",progname);
    exit(1);
  }
  arg = argv[1];
  dir = dirname(arg);
  arg = basename(arg,".hat");
  chdir(dir);
  HatFileSeq = openFile(arg,".hat");
  q_position = 0;
  err = q_fread(header,sizeof(char),8,HatFileSeq);
  if (err!=8) {
    fprintf(stderr,"%s (error): file %s/%s is too short\n",progname,dir,arg);
    exit(1);
  }
  if (strncmp(header,"Hat",3)) {
    fprintf(stderr,"%s (error): file %s in directory %s\n",progname,arg,dir);
    fprintf(stderr,"   does not appear to be a Hat archive.  Quitting.\n");
    exit(1);
  }
  if (strncmp(header+3,VERSION,4)) {
    fprintf(stderr,"%s (warning): file %s in directory %s\n",progname,arg,dir);
    fprintf(stderr,"   appears to be a Hat archive in format %s\n",header+3);
    fprintf(stderr,"   but this tool deals with format version %s\n",VERSION);
    fprintf(stderr,"   I'm continuing, but there may be unexpected errors.\n");
  }
  /* errorRoot*/ q_readFO();
  /* errorMsg */ q_readFO();

  observableInfo();

  exit(0);
}
#endif
